// ==============================================================
//	This file is part of Glest (www.glest.org)
//
//	Copyright (C) 2001-2005 Martio Figueroa
//
//	You can redistribute this code and/or modify it under 
//	the terms of the GNU General Public License as published 
//	by the Free Software Foundation; either version 2 of the 
//	License, or (at your option) any later version
// ==============================================================

#include "ai_rule.h"

#include "ai.h"
#include "ai_interface.h"
#include "unit.h"
#include "leak_dumper.h"

#include <algorithm>

using Shared::Graphics::Vec2i;

namespace Glest{ namespace Game{

// ===============================
// 	class AiRule  
// ===============================

AiRule::AiRule(Ai *ai){	
	this->ai= ai;
}

// ========================================
// 	class AiRuleWorkerHarvest 
// ========================================

AiRuleWorkerHarvest::AiRuleWorkerHarvest(Ai *ai): 
	AiRule(ai)
{
	stoppedWorkerIndex= -1;
}

bool AiRuleWorkerHarvest::test(){
	return ai->anyActingAble(&stoppedWorkerIndex, ccHarvest, scStop);
}

void AiRuleWorkerHarvest::execute(){
	ai->harvest(stoppedWorkerIndex);
}

// ========================================
// 	class AiRuleScoutPatrol
// ========================================

AiRuleScoutPatrol::AiRuleScoutPatrol(Ai *ai): 
	AiRule(ai)
{
}

bool AiRuleScoutPatrol::test(){
	return ai->isStableBase();
}

void AiRuleScoutPatrol::execute(){
	ai->sendScoutPatrol();
}

// ========================================
// 	class AiRuleRepair
// ========================================

AiRuleRepair::AiRuleRepair(Ai *ai): 
	AiRule(ai)
{
}

bool AiRuleRepair::test(){
	AiInterface *aiInterface= ai->getAiInterface();

	//look for a damaged unit
	for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
		const Unit *u= aiInterface->getMyUnit(i);
		if(u->getHpRatio()<1.f){
			damagedUnitIndex= i;
			return true;
		}
	}
	return false;
}

void AiRuleRepair::execute(){
	AiInterface *aiInterface= ai->getAiInterface();
	const Unit *damagedUnit= aiInterface->getMyUnit(damagedUnitIndex);
	
	//find a repairer and issue command
	for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
		const Unit *u= aiInterface->getMyUnit(i);
		const RepairCommandType *rct= static_cast<const RepairCommandType *>(u->getType()->getFirstCtOfClass(ccRepair));
		if(rct!=NULL){
			if(rct->isRepairableUnitType(damagedUnit->getType())){
				ai->getAiInterface()->giveCommand(i, rct, damagedUnit->getPos()); 
				return;
			}
		}
	}
}

// ========================================
// 	class AiRuleWarriorReturn
// ========================================

AiRuleWarriorReturn::AiRuleWarriorReturn(Ai *ai): 
	AiRule(ai)
{
	stoppedWarriorIndex= -1;
}

bool AiRuleWarriorReturn::test(){
	return ai->anyActingAble(&stoppedWarriorIndex, ccAttack, scStop);
}

void AiRuleWarriorReturn::execute(){
	ai->returnBase(stoppedWarriorIndex);
}

// ========================================
// 	class AiRuleMassiveAttack
// ========================================

AiRuleMassiveAttack::AiRuleMassiveAttack(Ai *ai): 
	AiRule(ai)
{
}

bool AiRuleMassiveAttack::test(){
	
	const int massiveRadius= 1000000;

	if(ai->isStableBase()){
		ultraAttack= false;
		return ai->beingAttacked(attackPos, field, massiveRadius);
	}
	else{
		ultraAttack= true;
		return ai->beingAttacked(attackPos, field, baseRadius);
	}   
}

void AiRuleMassiveAttack::execute(){
	ai->massiveAttack(attackPos, field, ultraAttack);
}

// ========================================
// 	class AiRuleAddTasks
// ========================================

AiRuleAddTasks::AiRuleAddTasks(Ai *ai):
	AiRule(ai)
{
}

bool AiRuleAddTasks::test(){
	return !ai->anyTask() || ai->getCountOfClass(ucWorker)<4;
}

void AiRuleAddTasks::execute(){
	int buildingCount= ai->getCountOfClass(ucBuilding);
	int warriorCount= ai->getCountOfClass(ucWarrior);
	int workerCount= ai->getCountOfClass(ucWorker);
	
	float buildingRatio= ai->getRatioOfClass(ucBuilding);
	float warriorRatio= ai->getRatioOfClass(ucWarrior);
	float workerRatio= ai->getRatioOfClass(ucWorker);

	//standard tasks

	//emergency workers
	if(workerCount<4){
		ai->addPriorityTask(new ProduceTask(ucWorker));
	}
	else{
		//buildings
		if(buildingCount<6) ai->addTask(new BuildTask());
		if(buildingRatio<0.20) ai->addTask(new BuildTask());
		
		//workers
		if(workerCount<5) ai->addTask(new ProduceTask(ucWorker));
		if(workerCount<10) ai->addTask(new ProduceTask(ucWorker));
		if(workerRatio<0.20) ai->addTask(new ProduceTask(ucWorker));
		if(workerRatio<0.30) ai->addTask(new ProduceTask(ucWorker));

		//warriors
		if(warriorCount<10) ai->addTask(new ProduceTask(ucWarrior));
		if(warriorRatio<0.20) ai->addTask(new ProduceTask(ucWarrior));
		if(warriorRatio<0.30) ai->addTask(new ProduceTask(ucWarrior));
		if(workerCount>=10) ai->addTask(new ProduceTask(ucWarrior));

		//upgrades
		if(ai->isStableBase()){
			ai->addTask(new UpgradeTask());
		}
	}
}

// ========================================
// 	class AiRuleBuildOneFarm
// ========================================

AiRuleBuildOneFarm::AiRuleBuildOneFarm(Ai *ai):
	AiRule(ai)
{
}

bool AiRuleBuildOneFarm::test(){
	AiInterface *aiInterface= ai->getAiInterface();

	//for all units
	for(int i=0; i<aiInterface->getMyFactionType()->getUnitTypeCount(); ++i){
		const UnitType *ut= aiInterface->getMyFactionType()->getUnitType(i);
		
		//for all production commands
		for(int j=0; j<ut->getCommandTypeCount(); ++j){
			const CommandType *ct= ut->getCommandType(j);
			if(ct->getClass()==ccProduce){
				const UnitType *producedType= static_cast<const ProduceCommandType*>(ct)->getProducedUnit();
				
				//for all resources
				for(int k=0; k<producedType->getCostCount(); ++k){
					const Resource *r= producedType->getCost(k);

					//find a food producer in the farm produced units
					if(r->getAmount()<0 && r->getType()->getClass()==rcConsumable && ai->getCountOfType(ut)==0){
						farm= ut;
						return true;
					}
				}
			}
		}
	}
	return false;
}

void AiRuleBuildOneFarm::execute(){
	ai->addPriorityTask(new BuildTask(farm));
}

// ========================================
// 	class AiRuleProduceResourceProducer
// ========================================

AiRuleProduceResourceProducer::AiRuleProduceResourceProducer(Ai *ai):
	AiRule(ai)
{
	interval= shortInterval;
}

bool AiRuleProduceResourceProducer::test(){
	//emergency tasks: resource buildings
	AiInterface *aiInterface= ai->getAiInterface();

	//consumables first
	for(int i=0; i<aiInterface->getTechTree()->getResourceTypeCount(); ++i){
        rt= aiInterface->getTechTree()->getResourceType(i);
		const Resource *r= aiInterface->getResource(rt);   
		if(rt->getClass()==rcConsumable && r->getBalance()<0){
			interval= longInterval;
			return true;
			
        }
    }

	//statics second
	for(int i=0; i<aiInterface->getTechTree()->getResourceTypeCount(); ++i){
        rt= aiInterface->getTechTree()->getResourceType(i);
		const Resource *r= aiInterface->getResource(rt);   
		if(rt->getClass()==rcStatic && r->getAmount()<minStaticResources){
			interval= longInterval;
			return true;
			
        }
    }

	interval= shortInterval;
	return false;
}

void AiRuleProduceResourceProducer::execute(){
	ai->addPriorityTask(new ProduceTask(rt));
	ai->addTask(new BuildTask(rt));
}

// ========================================
// 	class AiRulePerformTask
// ========================================

AiRulePerformTask::AiRulePerformTask(Ai *ai):
	AiRule(ai)
{
}

bool AiRulePerformTask::test(){
	return ai->anyTask();
}

void AiRulePerformTask::execute(){
	const Task *task= ai->getTask();
	if(task!=NULL){

		switch(task->getClass()){
		case tcProduce:
			produce(static_cast<const ProduceTask*>(task));		 
			break;
		
		case tcBuild:
			build(static_cast<const BuildTask*>(task));		 
			break;
		
		case tcUpgrade:
			upgrade(static_cast<const UpgradeTask*>(task));		 
			break;

		default:
			assert(false);
		}
	}
}

void AiRulePerformTask::produce(const ProduceTask *pt){
	
	//generic produce task, produce random unit that has the skill or produces the resource
	if(pt->getUnitType()==NULL){
		produceGeneric(pt);
	}

	//specific produce task, produce if possible, retry if not enough resources
	else{
		produceSpecific(pt);
	}

	//remove the task
	ai->removeTask(pt);
}

void AiRulePerformTask::produceGeneric(const ProduceTask *pt){
	typedef vector<const UnitType*> UnitTypes; 
	UnitTypes ableUnits;
	AiInterface *aiInterface= ai->getAiInterface();

	//for each unit, produce it if possible
	for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
		
		//for each command
		const UnitType *ut= aiInterface->getMyUnit(i)->getType();
		for(int j=0; j<ut->getCommandTypeCount(); ++j){
			const CommandType *ct= ut->getCommandType(j);
			
			//if the command is produce
			if(ct->getClass()==ccProduce || ct->getClass()==ccMorph){
				
				const UnitType *producedUnit= static_cast<const UnitType*>(ct->getProduced());
				bool produceIt= false;

				//if the unit produces the resource
				if(pt->getResourceType()!=NULL){
					const Resource *r= producedUnit->getCost(pt->getResourceType());
					if(r!=NULL && r->getAmount()<0){
						produceIt= true;
					}
				}

				else{
					//if the unit is from the right class
					if(producedUnit->isOfClass(pt->getUnitClass())){
						if(aiInterface->reqsOk(ct) && aiInterface->reqsOk(producedUnit)){
							produceIt= true;
						}
					}
				}

				if(produceIt){
					//if the unit is not already on the list
					if(find(ableUnits.begin(), ableUnits.end(), producedUnit)==ableUnits.end()){
						ableUnits.push_back(producedUnit);
					}
				}
			}
		}
	}

	//add specific produce task
	if(!ableUnits.empty()){

		//priority for non produced units
		for(int i=0; i<ableUnits.size(); ++i){
			if(ai->getCountOfType(ableUnits[i])==0){
				if(randRange(0, 1)==0){
					ai->addTask(new ProduceTask(ableUnits[i]));
					return;
				}
			}
		}

		//normal case
		ai->addTask(new ProduceTask(ableUnits[randRange(0, ableUnits.size()-1)]));
	}
}

void AiRulePerformTask::produceSpecific(const ProduceTask *pt){
	
	AiInterface *aiInterface= ai->getAiInterface();
	
	//if unit meets requirements
	if(aiInterface->reqsOk(pt->getUnitType())){
	
		//if unit doesnt meet resources retry
		if(!aiInterface->checkCosts(pt->getUnitType())){
			ai->retryTask(pt);
			return;
		}

		//produce specific unit
		vector<int> producers;
		const CommandType *defCt= NULL;

		//for each unit
		for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
			
			//for each command
			const UnitType *ut= aiInterface->getMyUnit(i)->getType();
			for(int j=0; j<ut->getCommandTypeCount(); ++j){
				const CommandType *ct= ut->getCommandType(j);
				
				//if the command is produce
				if(ct->getClass()==ccProduce || ct->getClass()==ccMorph){
					const UnitType *producedUnit= static_cast<const UnitType*>(ct->getProduced());
	
					//if units match
					if(producedUnit == pt->getUnitType()){
						if(aiInterface->reqsOk(ct)){
							defCt= ct;
							producers.push_back(i);
						}
					} 
				}
			}
		}

		//produce from random producer
		if(!producers.empty()){
			int producerIndex= producers[randRange(0, producers.size()-1)];
			aiInterface->giveCommand(producerIndex, defCt);
		}
	}
}

void AiRulePerformTask::build(const BuildTask *bt){
	
	//generic build task, build random building that can be built
	if(bt->getUnitType()==NULL){
		buildGeneric(bt);
	}
	//specific building task, build if possible, retry if not enough resources or not position
	else{
		buildSpecific(bt);
	}

	//remove the task
	ai->removeTask(bt);
}

void AiRulePerformTask::buildGeneric(const BuildTask *bt){
	//find buildings that can be built
	AiInterface *aiInterface= ai->getAiInterface();
	typedef vector<const UnitType*> UnitTypes; 
	UnitTypes buildings;
	
	//for each unit
	for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
		
		//for each command
		const UnitType *ut= aiInterface->getMyUnit(i)->getType();
		for(int j=0; j<ut->getCommandTypeCount(); ++j){
			const CommandType *ct= ut->getCommandType(j);
			
			//if the command is build
			if(ct->getClass()==ccBuild){
				const BuildCommandType *bct= static_cast<const BuildCommandType*>(ct);
				
				//for each building
				for(int k=0; k<bct->getBuildingCount(); ++k){
					const UnitType *building= bct->getBuilding(k);	
					if(aiInterface->reqsOk(bct) && aiInterface->reqsOk(building)){
						
						//if any building, or produces resource
						const ResourceType *rt= bt->getResourceType();
						const Resource *cost= building->getCost(rt);
						if(rt==NULL || (cost!=NULL && cost->getAmount()<0)){
							buildings.push_back(building);
						}
					}
				}
			}
		}
	}

	//add specific build task
	if(!buildings.empty()){

		//build the least built building
		bool buildingFound= false;
		for(int i=0; i<10 && !buildingFound; ++i){
			for(int j=0; j<buildings.size() && !buildingFound; ++j){
				const UnitType *building= buildings[j];
				if(ai->getCountOfType(building)==i){
					ai->addTask(new BuildTask(building));
					buildingFound= true;
				}
			}
		}
	}
}

void AiRulePerformTask::buildSpecific(const BuildTask *bt){
	AiInterface *aiInterface= ai->getAiInterface();
	//if reqs ok
	if(aiInterface->reqsOk(bt->getUnitType())){

		//retry if not enough resources
		if(!aiInterface->checkCosts(bt->getUnitType())){
			ai->retryTask(bt);
			return;
		}

		vector<int> builders;
		const BuildCommandType *defBct= NULL;
						
		//for each unit
		for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
			
			//if the unit is not going to build
			const Unit *u= aiInterface->getMyUnit(i);
			if(!u->anyCommand() || u->getCurrCommand()->getCommandType()->getClass()!=ccBuild){

				//for each command
				const UnitType *ut= aiInterface->getMyUnit(i)->getType();
				for(int j=0; j<ut->getCommandTypeCount(); ++j){
					const CommandType *ct= ut->getCommandType(j);
					
					//if the command is build
					if(ct->getClass()==ccBuild){
						const BuildCommandType *bct= static_cast<const BuildCommandType*>(ct);
						
						//for each building
						for(int k=0; k<bct->getBuildingCount(); ++k){
							const UnitType *building= bct->getBuilding(k);	
							
							//if building match
							if(bt->getUnitType()==building){
								if(aiInterface->reqsOk(bct)){
									builders.push_back(i);
									defBct= bct;
								}
							}
						}
					}
				}
			}
		}

		//use random builder to build
		if(!builders.empty()){
			int builderIndex= builders[randRange(0, builders.size()-1)];
			Vec2i pos;		

			//if free pos give command, else retry
			if(ai->findPosForBuilding(bt->getUnitType(), pos)){
				aiInterface->giveCommand(builderIndex, defBct, pos, bt->getUnitType());
			}
			else{
				ai->retryTask(bt);
				return;
			}
		}
	}
}

void AiRulePerformTask::upgrade(const UpgradeTask *upgt){

	//upgrade any upgrade
	if(upgt->getUpgradeType()==NULL){
		upgradeGeneric(upgt);
	}
	//upgrade specific upgrade
	else{
		upgradeSpecific(upgt);
	}

	//remove the task
	ai->removeTask(upgt);
}

void AiRulePerformTask::upgradeGeneric(const UpgradeTask *upgt){
	
	typedef vector<const UpgradeType*> UpgradeTypes; 
	AiInterface *aiInterface= ai->getAiInterface();
	
	//find upgrades that can be upgraded
	UpgradeTypes upgrades;
	
	//for each upgrade, upgrade it if possible
	for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
		
		//for each command
		const UnitType *ut= aiInterface->getMyUnit(i)->getType();
		for(int j=0; j<ut->getCommandTypeCount(); ++j){
			const CommandType *ct= ut->getCommandType(j);
			
			//if the command is upgrade
			if(ct->getClass()==ccUpgrade){
				const UpgradeCommandType *upgct= static_cast<const UpgradeCommandType*>(ct);
				const UpgradeType *upgrade= upgct->getProducedUpgrade();
				if(aiInterface->reqsOk(upgct)){
					upgrades.push_back(upgrade);
				}
			}
		}
	}

	//add specific upgrade task
	if(!upgrades.empty()){
		ai->addTask(new UpgradeTask(upgrades[randRange(0, upgrades.size()-1)]));
	}
}

void AiRulePerformTask::upgradeSpecific(const UpgradeTask *upgt){
	
	AiInterface *aiInterface= ai->getAiInterface();

	//if reqs ok
	if(aiInterface->reqsOk(upgt->getUpgradeType())){

		//if resources dont meet retry
		if(!aiInterface->checkCosts(upgt->getUpgradeType())){
			ai->retryTask(upgt);
			return;
		}

		//for each unit
		for(int i=0; i<aiInterface->getMyUnitCount(); ++i){
			
			//for each command
			const UnitType *ut= aiInterface->getMyUnit(i)->getType();
			for(int j=0; j<ut->getCommandTypeCount(); ++j){
				const CommandType *ct= ut->getCommandType(j);
				
				//if the command is upgrade
				if(ct->getClass()==ccUpgrade){
					const UpgradeCommandType *uct= static_cast<const UpgradeCommandType*>(ct);
					const UpgradeType *producedUpgrade= uct->getProducedUpgrade();
	
					//if upgrades match
					if(producedUpgrade == upgt->getUpgradeType()){
						if(aiInterface->reqsOk(uct)){
							aiInterface->giveCommand(i, uct);
						}
					} 
				}
			}
		}

	}
}

}}//end namespace 
