// ==============================================================
//	This file is part of Glest (www.glest.org)
//
//	Copyright (C) 2001-2005 Martio Figueroa
//
//	You can redistribute this code and/or modify it under 
//	the terms of the GNU General Public License as published 
//	by the Free Software Foundation; either version 2 of the 
//	License, or (at your option) any later version
// ==============================================================

#ifndef _GLEST_GAME_AIRULE_H_
#define _GLEST_GAME_AIRULE_H_

#include <string>

#include "vec.h"
#include "skill_type.h"

using std::string;

using Shared::Graphics::Vec2i;

namespace Glest{ namespace Game{

class Ai;
class Unit;
class UnitType;
class ProduceTask;
class BuildTask;
class UpgradeTask;
class ResourceType;

// ==============================================================
//	class AiRule  
//
///	An action that the AI will perform periodocally 
/// if the test succeeds
// ==============================================================

class AiRule{
protected:
	Ai *ai;

public:
	AiRule(Ai *ai);

	virtual int getTestInterval() const= 0;	//in milliseconds
	virtual string getName() const= 0;

	virtual bool test()= 0;
	virtual void execute()= 0;
};

// ========================================
// 	class AiRuleWorkerHarvest 
// ========================================

class AiRuleWorkerHarvest: public AiRule{
private:
	int stoppedWorkerIndex;

public:
	AiRuleWorkerHarvest(Ai *ai);
	
	virtual int getTestInterval() const	{return 2000;}
	virtual string getName() const		{return "Worker stopped => Order worker to harvest";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleScoutPatrol
// ========================================

class AiRuleScoutPatrol: public AiRule{
public:
	AiRuleScoutPatrol(Ai *ai);
	
	virtual int getTestInterval() const	{return 10000;}
	virtual string getName() const		{return "Base is stable => Send scout patrol";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleRepair
// ========================================

class AiRuleRepair: public AiRule{
private:
	int damagedUnitIndex;

public:
	AiRuleRepair(Ai *ai);
	
	virtual int getTestInterval() const	{return 10000;}
	virtual string getName() const		{return "Building Damaged => Repair";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleWarriorReturn
// ========================================

class AiRuleWarriorReturn: public AiRule{
private:
	int stoppedWarriorIndex;
public:
	AiRuleWarriorReturn(Ai *ai);
	
	virtual int getTestInterval() const	{return 5000;}
	virtual string getName() const		{return "Stopped warrior => Order return base";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleMassiveAttack
// ========================================

class AiRuleMassiveAttack: public AiRule{
private:
	static const int baseRadius= 25;

private:
	Vec2i attackPos;
	Field field;
	bool ultraAttack;

public:
	AiRuleMassiveAttack(Ai *ai);
	
	virtual int getTestInterval() const	{return 1000;}
	virtual string getName() const		{return "Unit under attack => Order massive attack";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleAddTasks
// ========================================

class AiRuleAddTasks: public AiRule{
public:
	AiRuleAddTasks(Ai *ai);
	
	virtual int getTestInterval() const	{return 5000;}
	virtual string getName() const		{return "Tasks empty => Add tasks";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleBuildOneFarm
// ========================================

class AiRuleBuildOneFarm: public AiRule{
private:
	const UnitType *farm;

public:
	AiRuleBuildOneFarm(Ai *ai);

	virtual int getTestInterval() const	{return 10000;}
	virtual string getName() const		{return "No farms => Build one";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRuleProduceResourceProducer
// ========================================

class AiRuleProduceResourceProducer: public AiRule{
private:
	static const int minStaticResources= 20;
	static const int longInterval=	60000;
	static const int shortInterval= 5000;
	const ResourceType *rt;
	int interval;

public:
	AiRuleProduceResourceProducer(Ai *ai);
	
	virtual int getTestInterval() const	{return interval;}
	virtual string getName() const		{return "No resources => Build Resource Producer";}

	virtual bool test();
	virtual void execute();
};

// ========================================
// 	class AiRulePerformTasks
// ========================================

class AiRulePerformTask: public AiRule{
public:
	AiRulePerformTask(Ai *ai);
	
	virtual int getTestInterval() const	{return 2000;}
	virtual string getName() const		{return "Tasks not empty => Perform tasks";}

	virtual bool test();
	virtual void execute();

private:
	void produce(const ProduceTask *pt);
	void produceGeneric(const ProduceTask *pt);
	void produceSpecific(const ProduceTask *pt);

	void build(const BuildTask *bt);
	void buildGeneric(const BuildTask *bt);
	void buildSpecific(const BuildTask *bt);

	void upgrade(const UpgradeTask *upgt);
	void upgradeSpecific(const UpgradeTask *upgt);
	void upgradeGeneric(const UpgradeTask *upgt);
};

}}//end namespace 

#endif
