// ==============================================================
//	This file is part of Glest (www.glest.org)
//
//	Copyright (C) 2001-2005 Martio Figueroa
//
//	You can redistribute this code and/or modify it under 
//	the terms of the GNU General Public License as published 
//	by the Free Software Foundation; either version 2 of the 
//	License, or (at your option) any later version
// ==============================================================

#ifndef _GLEST_GAME_WORLD_H_
#define _GLEST_GAME_WORLD_H_

#include "vec.h"
#include "math_util.h"
#include "resource.h"
#include "tech_tree.h"
#include "tileset.h"
#include "console.h"
#include "map.h"
#include "minimap.h"
#include "logger.h"
#include "stats.h"
#include "time_flow.h"
#include "upgrade.h"
#include "water_effects.h"
#include "faction.h"
#include "unit_updater.h"

namespace Glest{ namespace Game{

using Shared::Graphics::Quad2i;
using Shared::Graphics::Rect2i;

class Faction;
class Unit;
class Config;
class Game;

const int maxPlayers= 4;

struct GameSettings{
     string mapPath;
     string tilesetPath;
     string techPath;
     int thisFactionIndex;
     int factionCount;
	 string factionTypesNames[maxPlayers]; //faction names
     int team[maxPlayers];
	 ControlerType factionControls[maxPlayers];
};

// ==============================================================
// 	class World
//
///	The game world: Map + Tileset + TechTree
// ==============================================================

class World{
public:
	static const int noTeam= 0;
	static const int generationArea= 100;
	static const float airHeight;
	static const int indirectSightRange= 5;

private:
	//config
	int updateFps;
	bool fogOfWar;
	bool fogOfWarSmoothing;
	int fogOfWarSmoothingFrameSkip;
	
	int thisFactionIndex;
	int updateCount;
	int thisFrameUpdateCount;

	Map map;
	Tileset tileset;
	TechTree techTree;
	TimeFlow timeFlow;

	UnitUpdater unitUpdater;
    WaterEffects waterEffects;
	Minimap minimap;
    Stats *stats;	//BattleEnd will delete this object

	vector<Faction*> factions;

public:
	World();
	void end(); //to die before selection does

	//get			
	int getMaxPlayers() const						{return maxPlayers;}
	int getThisFactionIndex() const					{return thisFactionIndex;}
	Faction *getThisFaction() const					{return factions[thisFactionIndex];}
	int getFactionCount() const						{return factions.size();}
	const Map *getMap() const 						{return &map;}
	const Tileset *getTileset() const 				{return &tileset;}
	const TechTree *getTechTree() const 			{return &techTree;}
	const TimeFlow *getTimeFlow() const				{return &timeFlow;}
	Tileset *getTileset() 							{return &tileset;}
	Map *getMap() 									{return &map;}
	Faction *getFaction(int i) const				{return factions[i];}
	const Minimap *getMinimap() const				{return &minimap;}
	Stats *getStats() const							{return stats;};
	const WaterEffects *getWaterEffects() const		{return &waterEffects;}

	//init & load
	void init(Game *game);
	void loadTileset(const string &dir);
	void loadTech(const string &dir);
	void loadMap(const string &path);
	
	//misc
	void update();
	void tick();
	bool placeUnit(const Vec2i &startLoc, int radius, Unit *unit, bool spaciated= false);
	void moveUnitCells(Unit *unit);
	bool toRenderUnit(const Unit *unit, const Quad2i &visibleQuad) const;
	bool toRenderUnit(const Unit *unit) const;
	Unit *nearestStore(const Vec2i &pos, int factionIndex, const ResourceType *rt);
	void computeFow();

private:
	//private init
	void initCells();
	void initSplattedTextures();
	void initFactionTypes(GameSettings *gs);
	void initMinimap();
	void initUnits();
	void initMap();
	void initExplorationState();
	
	//exploration
	void exploreCells(const Vec2i &newPos, int sightRange, int factionIndex);
};

}}//end namespace

#endif
