// ==============================================================
//	This file is part of Glest Shared Library (www.glest.org)
//
//	Copyright (C) 2001-2005 Martio Figueroa
//
//	You can redistribute this code and/or modify it under 
//	the terms of the GNU General Public License as published 
//	by the Free Software Foundation; either version 2 of the 
//	License, or (at your option) any later version
// ==============================================================

#include "particle_renderer_gl.h"

#include "opengl.h"
#include "texture_gl.h"
#include "model_renderer.h"
#include "math_util.h"
#include "leak_dumper.h"

namespace Shared{ namespace Graphics{ namespace Gl{

// ===========================================================================
//  ParticleRendererGl
// ===========================================================================


// ================ PUBLIC ========================================

ParticleRendererGl::ParticleRendererGl(){
	rendering= false;
}

void ParticleRendererGl::renderManager(ParticleManager *pm, ModelRenderer *mr){
	
	//assertions
	assertGl();

	//push state
	glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT  | GL_DEPTH_BUFFER_BIT | GL_POLYGON_BIT | GL_CURRENT_BIT | GL_LINE_BIT);
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

	//init state
	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glDisable(GL_STENCIL_TEST);
	glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//render
	rendering= true;
	pm->render(this, mr);
	rendering= false;

	//pop state
	glPopClientAttrib();
	glPopAttrib();

	//assertions
	assertGl();
}

void ParticleRendererGl::renderSystem(ParticleSystem *ps){
	//render particles
	renderSystemQuad(ps);
}

void ParticleRendererGl::renderSystemLine(ParticleSystem *ps){
	
	assertGl();
	assert(rendering);

	if(!ps->isEmpty()){
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_ALPHA_TEST);

		const Particle *p= ps->getParticle(0);
		glLineWidth(p->getSize());
		glBegin(GL_LINES);
		for(int i=0; i<ps->getAliveParticleCount(); ++i){
			p= ps->getParticle(i);
			glColor4fv(p->getColor().ptr());
			glVertex3fv(p->getPos().ptr());
			glVertex3fv(p->getLastPos().ptr());
		}
		glEnd();
	}

	assertGl();
}

void ParticleRendererGl::renderSystemLineAlpha(ParticleSystem *ps){
	
	assertGl();
	assert(rendering);

	if(!ps->isEmpty()){
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_ALPHA_TEST);

		const Particle *p= ps->getParticle(0);
		glLineWidth(p->getSize());
		glBegin(GL_LINES);
		for(int i=0; i<ps->getAliveParticleCount(); ++i){
			p= ps->getParticle(i);
			Vec4f color1= p->getColor();
			Vec4f color2= color1;
			color2.w= 0.0f;
			glColor4fv(color1.ptr());
			glVertex3fv(p->getPos().ptr());
			glColor4fv(color2.ptr());
			glVertex3fv(p->getLastPos().ptr());
		}
		glEnd();
	}

	assertGl();
}

void ParticleRendererGl::renderSingleModel(AttackParticleSystem *ps, ModelRenderer *mr){
	//render model
	if(ps->getModel()!=NULL){
		
		//init
		glEnable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
		glColor3f(1.f, 1.f, 1.f);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		
		//translate
		Vec3f pos= ps->getPos();
		glTranslatef(pos.x, pos.y, pos.z);

		//rotate
		Vec3f direction= ps->getDirection();
		Vec3f flatDirection= Vec3f(direction.x, 0.f, direction.z);
		Vec3f rotVector= Vec3f(0.f, 1.f, 0.f).cross(flatDirection);

		float angleV= radToDeg(atan2(flatDirection.length(), direction.y)) - 90.f;
		glRotatef(angleV, rotVector.x, rotVector.y, rotVector.z);
		float angleH= radToDeg(atan2(direction.x, direction.z));
		glRotatef(angleH, 0.f, 1.f, 0.f);
		
		//render
		mr->begin(true, true, false);
		mr->render(ps->getModel());
		mr->end();

		//end
		glPopMatrix();
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
	}
}


// ============== PRIVATE =====================================

void ParticleRendererGl::renderSystemQuad(const ParticleSystem *ps){

	assertGl();
	assert(rendering);

	if(ps->getTexture()!=NULL){
		glBindTexture(GL_TEXTURE_2D, static_cast<Texture2DGl*>(ps->getTexture())->getHandle());
	}
	else{
		glBindTexture(GL_TEXTURE_2D, 0);
	}
	glDisable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.0f);
	glEnable(GL_TEXTURE_2D);

	for(int i=0; i<ps->getAliveParticleCount(); ++i){
		renderQuad(ps->getParticle(i));
	}

	assertGl();
}

/*void ParticleRendererGl::renderSystemPoint(ParticleSystem *ps){
	
	assertGl();
	assert(rendering);

	if(!ps->isEmpty()){
		glBindTexture(GL_TEXTURE_2D, static_cast<Texture2DGl*>(ps->getTexture())->getHandle());
		glTexEnvf(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE);

		glEnable(GL_POINT_SPRITE_ARB);
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GREATER, 0.0f);
		glEnable(GL_TEXTURE_2D);

		glPointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB, Vec3f(0.f, 0.f, 1.f/60.f).ptr());

		Particle *p= ps->getParticle(0);
		glPointSize(p->getSize()*100);
		glBegin(GL_POINTS);
		for(int i=0; i<ps->getAliveParticleCount(); ++i){
			p= ps->getParticle(i);
			glColor4fv(p->getColor().ptr());
			glVertex3fv(p->getPos().ptr());
		}
		glEnd();
	}

	assertGl();
}*/

void ParticleRendererGl::renderQuad(const Particle *p){

	float size= p->getSize()/2.0f;

	glPushMatrix();
	glColor4fv(p->getColor().ptr());

	glTranslatef(p->getPos().x, p->getPos().y, p->getPos().z); 
	billBoard();

	glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(-size, size, 0.0f);
	
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-size, -size, 0.0f);

		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(size, size, 0.0f);

		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(size, -size, 0.0f);

	glEnd();

	glPopMatrix();
}

void ParticleRendererGl::billBoard(){

	float modelview[16];

	// get the current modelview matrix
	glGetFloatv(GL_MODELVIEW_MATRIX , modelview);

	// undo all rotations
	// beware all scaling is lost as well 
	modelview[0]=1.0f;
	modelview[1]=0.0f;
	modelview[2]=0.0f;
	modelview[4]=0.0f;
	modelview[5]=1.0f;
	modelview[6]=0.0f;
	modelview[8]=0.0f;
	modelview[9]=0.0f;
	modelview[10]=1.0f;

	// set the modelview with no rotations
	glLoadMatrixf(modelview);
}

}}} //end namespace
